	list	p=16f877A	; list directive to define processor
	include	<p16F877A.inc>
    
	errorlevel -302,-207	; suppress message 302 from list file
	errorlevel -305,-207	; suppress message 302 from list file

	__config _CP_OFF & _WDT_OFF & _BODEN_OFF & _PWRTE_ON & _XT_OSC & _WRT_OFF & _LVP_OFF & _CPD_OFF

;===========================================================================  
;
;     PIC16F877-A   
;                                 __________  
;                 +5v------!MCLR |1       40| RB7-I---- (KEYPAD  COL4)  
;            (Unused)------  RA0 |2       39| RB6-I---- (KEYPAD  COL3)
;            (Unused)------  RA1 |3       38| RB5-I---- (KEYPAD  COL2)
;            (Unused)------  RA2 |4       37| RB4-I---- (KEYPAD  COL1)
;            (Unused)------  RA3 |5       36| RB3  ---- (Unused)  
;            (Unused)------  RA4 |6       35| RB2  ---- (Unused)  
;            (Unused)------  RA5 |7       34| RB1  ---- (Unused)                  
;            (Unused)------  RE0 |8       33| RB0/INT-- (Unused)  
;            (Unused)------  RE1 |9       32| VDD  ---- +5v  
;            (Unused)------  RE2 |10      31| VSS  ---- GND  
;                 +5v------  VDD |11      30| RD7-O---- (KEYPAD  ROW4)   
;                 GND------  VSS |12      29| RD6-O---- (KEYPAD  ROW3)
;                XTAL------ OSC1 |13      28| RD5-O---- (KEYPAD  ROW2)  
;                XTAL------ OSC2 |14      27| RD4-O---- (KEYPAD  ROW1) 
;             (Unused)-----  RC0 |15      26| RC7-O---- (LCD pin 14)  LCD_DB7    
; LCD_rs (LCD pin 4) ------O-RC1 |16      25| RC6-O---- (LCD pin 13)  LCD_DB6  
; LCD_rw (LCD pin 5) ------O-RC2 |17      24| RC5-O---- (LCD pin 12)  LCD_DB5  
; LCD_e  (LCD pin 6) ------O-RC3 |18      23| RC4-O---- (LCD pin 11)  LCD_DB4  
;            (Unused)------  RD0 |19      22| RD3  ---- (Unused)  
;            (Unused)------  RD1 |20      21| RD2  ---- (Unused)  
;                                 ----------  
;
;	LCD (Hitachi)  
;                                 ________  
;    		                 VSS |1       |   
;                            VDD |2       |   
;                 CONTRAST   VEE |3       |   
;                 RC1------  RS  |4       |   
;                 RC2------  R/W |5       |   
;                 RC3------  E   |6       |   
;                       (Unused) |7       |                   
;                       (Unused) |8       |   
;                       (Unused) |9       |   
;                       (Unused) |10      |  
;                 RC4------  DB4 |11      |    
;                 RC5------  DB5 |12      |   
;                 RC6------  DB6 |13      |      
;                 RC7------  DB7 |14      |     
;          BACKLIGHT POWER   VDD |15      |     
;                   GROUND   VSS |16      |   
;                                 --------  
;
;
;    KEYPAD
;                     RB4        RB5         RB6          RB7
;                      |          |           |            |
;                      |          |           |            |
;                      |          |           |            |
;
;                    COL 1       COL 2       COL 3        COL 4   
;                 +---------+  +--------+  +--------+   +--------+  
; RD4-------ROW 1 |    1    |  |   2    |  |   3    |   |   A    |
;                 +---------+  +--------+  +--------+   +--------+ 
;                 +---------+  +--------+  +--------+   +--------+  
; RD4-------ROW 2 |    4    |  |   5    |  |   6    |   |   B    |
;                 +---------+  +--------+  +--------+   +--------+ 
;                 +---------+  +--------+  +--------+   +--------+  
; RD4-------ROW 3 |    7    |  |   8    |  |   9    |   |   C    |
;                 +---------+  +--------+  +--------+   +--------+  
;                 +---------+  +--------+  +--------+   +--------+  
; RD4-------ROW 4 |    *    |  |   0    |  |   #    |   |   D    |
;                 +---------+  +--------+  +--------+   +--------+  
;
;
;=============================================================================
; DECLARATIONS
;=============================================================================
RESET_V		equ	0X0000			; ADDRESS OF RESET VECTOR.
ISR_V		equ	0X0004			; ADDRESS OF INTERRUPT VECTOR.
OSC_FREQ 	equ D'4000000'		; OSCILLATOR FREQUENCY IS 4 MHZ.

LCD_RS		equ	1				;0 = COMMANDS. 1 = DATA.
LCD_RW		equ	2				;0 = WRITE TO LCD. 1 = READ LCD.
LCD_E		equ	3				;ENABLE THE LCD TO ACCEPT INPUT. PULSE THIS.
LCDBUSY		equ	7				;WHEN THE LCD IS BUSY RC7 IS SET HIGH.


; PORTB BITS
RB7			equ	7				;KEYBOARD, COL 4.
RB6			equ	6				;KEYBOARD, COL 3.
RB5			equ	5 				;KEYBOARD, COL 2.
RB4			equ	4				;KEYBOARD, COL 1.

; PORTD BITS
RD7			equ	7				;KEYBOARD, ROW 1.
RD6			equ	6				;KEYBOARD, ROW 2.
RD5			equ	5				;KEYBOARD, ROW 3.
RD4			equ	4				;KEYBOARD, ROW 4.

RD3			equ	3				;KEY CODE, BIT 3, USE 1K + LED TO VSS.
RD2			equ	2				;KEY CODE, BIT 2, USE 1K + LED TO VSS.
RD1			equ	1				;KEY CODE, BIT 1, USE 1K + LED TO VSS.
RD0			equ	0				;KEY CODE, BIT 0, USE 1K + LED TO VSS.

KEYS_BIT	equ	0				;BIT FOR SETTING THE KEYBOARD FLAG.
MENU_BIT	equ 0				;BIT FOR USER BEING IN MENU.
INT_BIT		equ 0
		

; REGISTERS.
		cblock  H'20'       ; NOTE: START DATA BLOCK.
;FREQUENCY "ENTRY" VALUES.
POS_10M	
POS_1M
POS_100K
POS_10K
POS_1K
POS_100
POS_10
POS_1						;** FCHAR_STRT MUST BE SET TO THIS ADDRESS **

FREQ_10M	
FREQ_1M
FREQ_100K
FREQ_10K
FREQ_1K
FREQ_100
FREQ_10
FREQ_1						;** FCHAR_FREQ MUST BE SET TO THIS ADDRESS **

FCHAR_MAX					;THIS IS THE TRUE COUNT OF CHARS ENTERED.
FCHAR_MAX_TEMP				;THIS IS THE TRUE COUNT OF OLD FREQ ENTERED.
FCHAR_CNT					;USED EVERYTIME WE SHIFT CHARS (UPDATED).
FCHAR_CNT1					;USED WHEN TRANSFERING FROM POS_XX TO FREQ_XX
FCHAR_STRT					;START OF POS_1 IN ADDRESS LOCATION...
FCHAR_TEMP					;START OF TEMP_1 IN ADDRESS LOCATION...
FCHAR_FREQ					;START OF FREQ_1 IN ADDRESS LOCATION...

POS1
POS2

;ENCODER STEP VARIABLE.
STEP_VAL

;EEPROM VARIABLES.
ADDR
VALUE

;KEYPAD VARIABLES.       		
TEMP						;TEMP STORAGE.
LAST_KEY					;LAST KEY PRESSED.
LAST_CHAR					;LAST KEY PRESSED.

		
TIMER1						; USED IN DELAY ROUTINES
TIMER2						;   " 

;LCD VARIABLES.
LCD_CHAR					;THE BYTE REPRESENTING THE LCD CHARACTER. 
LCD_READ 					;VARIABLE CONTAINING THE READ FOR LCD_BUSY. 
LCD_VALUE 					;THE LCD RS LINE - FLAG VALUE. 

;INTERRUPT FLAG VARIABLE.
INTERRUPT_FLAG

;OTHER FLAG VARIABLES.
EEPROM_FLAG
NEW_FREQ_FLAG
MENU_FLAG

MENU_COUNT					;USED TO INCREMENT UPON DISPLAYING NEW MENU ITEM.
		
;INTERRUPT TEMPORARY SWAP OUT VARIABLES.
W_TEMP
STATUS_TEMP
PCLATH_TEMP

;DELAY VARIABLES.
DELAY_VAL					; USED IN DELAY ROUTINES.
X_DLY_VAL					; "
COUNT						; USED AS G.P. COUNTER.

		endc				; NOTE: END OF DATA BLOCK.    
   
;=============================================================================
; PROGRAM START.
;***************
;
; NOTE: - 	Programs always start at ORG 0x00. The reset vector always starts
;			at 0x04 so four lines down from 00. This means if you are 
;			including an interupt routine you must jump over 0x04 or your 
;			will run into it.		
;=============================================================================
		org		RESET_V		; RESET vector location
		reset	GOTO MAIN	;JUMPING TO AVOID THE INTERRUPT VECTOR.

;=============================================================================
;THIS IS THE INTERUPT ROUTINE.
;
;PURPOSE: - To interrupt the microcontroller inorder to perform the following 
;			tasks:
;				- what interrupt was it??
;				- Set the approriate flag.
;			Its important to get in and get out of the interrupt routine
;			effectively... so we set flags rather than do any work.
;
;CODE 
;SOURCE: - 	Much of this routine was recovered from MICROCHIP'S DATASHEET. See
;			page ?? where it has examples of the INTERRUPT routine for and
;			how to save off the special registers and then after recover them. 
;=============================================================================
INTERRUPT	org	ISR_V		;INTERRUPT VECTOR LOCATION.

		movwf	W_TEMP 		;COPY W TO TEMP REGISTER
		swapf	STATUS,	W 	;SWAP STATUS TO BE SAVED INTO W
		clrf	STATUS 		;BANK 0, REGARDLESS OF CURRENT BANK, CLEARS IRP,RP1,RP0
		movwf	STATUS_TEMP ;SAVE STATUS TO BANK ZERO STATUS_TEMP REGISTER
		movf	PCLATH, W 	;ONLY REQUIRED IF USING PAGES 1, 2 AND/OR 3
		movwf	PCLATH_TEMP ;SAVE PCLATH INTO W
		clrf	PCLATH 		;PAGE ZERO, REGARDLESS OF CURRENT PAGE

		call	WAIT_15MS

		movlw   B'11110000'	;CHECK FOR KEY DOWN.
		movwf   TEMP  
		movf    PORTB,	W
		subwf   TEMP,	F
		btfsc	STATUS, Z
		goto    DONE		;ALL KEYS UP SO RETURN.
		
		bsf		INTERRUPT_FLAG,	KEYS_BIT
		
DONE
		clrf    PORTB		;CLEAR PORT B.
		clrf    PORTD		;CLEAR PORT D.
		movf    PORTB,	F

 		bcf     INTCON,RBIF;CLEAR Port B interrupt FLAG

		movf	PCLATH_TEMP, W ;RESTORE PCLATH
		movwf	PCLATH 		;MOVE W INTO PCLATH
		swapf	STATUS_TEMP,W ;SWAP STATUS_TEMP REGISTER INTO W
							;(SETS BANK TO ORIGINAL STATE)
		movwf	STATUS 		;MOVE W INTO STATUS REGISTER
		swapf 	W_TEMP,F 	;SWAP W_TEMP
		swapf 	W_TEMP,W 	;Swap W_TEMP into W

		retfie

;=============================================================================
;INCLUDE FILES HERE.
;*******************
;=============================================================================
		include "KEYPAD_ROUTINES.INC"
		include "LCD_ROUTINES.INC"
		;include "LCD_LABELS.INC"

;=============================================================================
;"MAIN" STARTS HERE.
;*******************
;
;PURPOSE: - Here we call the PIC initialisation routine and the LCD 
;			initialisation routine... display an opeing message, we set our
;			pivotal variables and call our start routine.
;=============================================================================	
MAIN
		call	WAIT_15MS	;GIVE THE MICRO A CHANCE TO SETTLE DOWN.
		call	PIC_INIT	;INITIATE THE PIC - SET PORTS ENABLE WEAK PULLUPS.
		;call	LCD_INIT	;INITIATE THE LCD NEXT (LCD_ROUTINES.INC).



							;NOTE - We construct our frequency using eight
							; bytes. The first is situated at the address
							; H'27' and goes UP ie singles at H'27' tens
							; at H'26' and so on. We need to set our 
							; variables first and  this includes the first
							; just mentioned - H'27' for our location just
							; temporary location - H'37' for our frequency 
							; location.

							; FCHAR_MAX is PIVOTAL in this program.
							; It tells us how many numbers have been 
							; entered.
						
		movlw	H'27'
		movwf	FCHAR_STRT	;SO SET OUR FIRST "POS" OR TEMPORARY LOCATION.

		movlw	H'2F'
		movwf	FCHAR_FREQ	;NOW SET OUR "FREQ" LOCATION.

		movlw	H'00'
		movwf	FCHAR_MAX	;ALWAYS START AT "0" FOR OUR ENTER COUNT.
		
		movlw	H'00'
		movwf	INTERRUPT_FLAG ; OUR INTERRUPT FLAG MUST BE CLEARED.

		movlw	H'00'
		movwf	NEW_FREQ_FLAG ; NOW CLEAR OUR FREQUENCY ENTER FLAG.

		movlw	H'00'
		movwf	MENU_FLAG 	; NOW CLEAR OUR FREQUENCY ENTER FLAG.

		bcf		NEW_FREQ_FLAG,	INT_BIT	; CLEAR OUR EEPROM FLAG.

							;NOTE: - This next routine goes off first to
							; read the eerpom, check if it has anything
							; in it, then based on if it has or hasn't
							; set the EEPROM_FLAG accordingly. By the way
							; if the eeprom has a frequency it loads
							; this into the FCHAR_FREQ registers ready
							; for the LCD to display them...

		goto MENU ; OUR FIRST ROUTINE. 

		
		;=============================================================================
;3.11 RUN THROUGH THE MENU OPTIONS.
;
;PURPOSE: -  If "D" has been entered then the user wants to go to the menu.
;			The menu is simply the "other" options displayed in a never
;			ending loop. The user can break the loop by pressing the "#" key.
;
;			This is a never ending loop that the user can break in on and 
;			select an option.  Anyone can add to this if they want to...
;							   *****************************************
;=============================================================================
MENU
		bcf		INTERRUPT_FLAG,	KEYS_BIT ; CLEAR OUR INTERRUPT FLAG FOR KEYS.
		bsf		MENU_FLAG, MENU_BIT	; SET THIS FLAG TO TEST FOR BEING IN MENU.

		call	 PRESS_#	;TELL USER TO PRESS # FOR ENTER...

						;NOTE - if "D" is pressed then the program goes into 
						; an indefinite loop for the user to select an option  
						; of exit...

		movlw	H'01'
		movwf	MENU_COUNT

MENU_LOOP
        call    RESET_STARTUP_FREQ_LCD ; GOT THE SET STARTUP FREQ ROUTINE.
		incf	MENU_COUNT
        call    WAIT_256MS
        ;call    WAIT_500MS

		btfsc	INTERRUPT_FLAG,KEYS_BIT	; INTERRUPT FLAG BEEN SET?
		call	MENU_INTERRUPT	;YES. SO GO FIND OUT WHAT KEY AND PROCESS....
	;	bcf		INTERRUPT_FLAG,	KEYS_BIT ; CLEAR OUR INTERRUPT FLAG FOR KEYS.

		btfss	MENU_FLAG, MENU_BIT
		goto	EXIT_MENU_LOOP

        call    CALIBRATE_LCD	;GOTO THE CALIBRATE ROUTINE.
		incf	MENU_COUNT
        call    WAIT_256MS
       ; call    WAIT_500MS

		btfsc	INTERRUPT_FLAG,KEYS_BIT	; INTERRUPT FLAG BEEN SET?
		call	MENU_INTERRUPT	;YES. SO GO FIND OUT WHAT KEY AND PROCESS....
 ;		bcf		INTERRUPT_FLAG,	KEYS_BIT ; CLEAR OUR INTERRUPT FLAG FOR KEYS.

		btfss	MENU_FLAG, MENU_BIT
		goto	EXIT_MENU_LOOP

        call    MENU_EXIT_LCD	;EXIT
		incf	MENU_COUNT
        call    WAIT_256MS
        ;call    WAIT_500MS

		btfsc	INTERRUPT_FLAG,KEYS_BIT	; INTERRUPT FLAG BEEN SET?
		call	MENU_INTERRUPT	;YES. SO GO FIND OUT WHAT KEY AND PROCESS....
;		bcf		INTERRUPT_FLAG,	KEYS_BIT ; CLEAR OUR INTERRUPT FLAG FOR KEYS.

		btfss	MENU_FLAG, MENU_BIT
		goto	EXIT_MENU_LOOP

		movlw	H'00'
		movwf	MENU_COUNT

        goto    MENU_LOOP	; Repeat forever

EXIT_MENU_LOOP

		bcf		MENU_FLAG, MENU_BIT	; SET THIS FLAG TO TEST FOR BEING IN MENU.
		return

;=============================================================================
;
;=============================================================================
MENU_INTERRUPT
		call	GETKEY		;GETKEY IS IN THE KEYPAD INCLUDE FILE.

		sublw	D'11'	
		btfsc	STATUS,Z	;IS THE ZERO BIT SET? DID THEY PRESS "#"?
		goto	TEST_FLAG	;YES. SO TEST THE FLAG.		
		goto	WRONG_KEY	;NO. SO CLEAR FLAGS AND RETURN.


TEST_FLAG
		movf	MENU_COUNT,W	;MOVE OUR CHAR INTO "W".
		sublw	D'01'	
		btfsc	STATUS,Z	;IS THE ZERO BIT SET?
		goto	RESET_STARTUP_FREQ ; USER WANTS TO CHANGE EEPROM FREQ.

		movf	MENU_COUNT,W	;MOVE OUR CHAR INTO "W".
		sublw	D'02'	
		btfsc	STATUS,Z	;IS THE ZERO BIT SET?
		goto	CALIBRATE 	;USER WANTS TO CHANGE EEPROM FREQ.

		movf	MENU_COUNT,W	;MOVE OUR CHAR INTO "W".
		sublw	D'03'	
		btfsc	STATUS,Z	;IS THE ZERO BIT SET?
		goto	MENU_EXIT	; USER WANTS TO CHANGE EEPROM FREQ.

WRONG_KEY
		bcf		INTERRUPT_FLAG,	KEYS_BIT ;WRONG KEY PRESSED SO RESET
							;			  THE INTERRUPT BIT.	
			
		return				;(RETURN TO LOOP FROM THIS ROUTINE.)





;=============================================================================
;3.12 USER WANTS TO ADD/CHANGE THE START UP FREQUENCY LOADED IN THE EEPROM.
;=============================================================================
RESET_STARTUP_FREQ
		bcf		MENU_FLAG, MENU_BIT	; CLR THIS FLAG TO TEST FOR BEING IN MENU.

		call	LOAD_NEW_EE_FREQ ; SEE 1.2 ABOVE. SETS UP FOR EERPOM LOAD.

		return				;(AND RETURN TO LOOP HERE)

;=============================================================================
;3.13 USER WANTS CALIBRATE THE FUNCTION GENERATOR.
;=============================================================================
CALIBRATE
		bcf		MENU_FLAG, MENU_BIT	; CLR THIS FLAG TO TEST FOR BEING IN MENU.

		return				;(AND RETURN TO LOOP HERE)

;=============================================================================
;3.14 USER WANTS EXIT THE MENU LOOP.
;=============================================================================
MENU_EXIT

		bcf		MENU_FLAG, MENU_BIT	; CLR THIS FLAG TO TEST FOR BEING IN MENU.

		return				;(AND RETURN TO LOOP HERE)

		
		
		
		
		
		
		
		
		
		
		
		
		
		;===========================================================================
;DELAY ROUTINES.
;***************
;
; DELAY_TIME	= ((DELAY_value * 3) + 4) * Cycle_time
; DELAY_VALUE	= (Delay_time - (4 * Cycle_time)) / (3 * Cycle_time)
;
; i.e. (@ 4MHz crystal)
; Delay_time	= ((32 * 3) + 4) * 1uSec
;		= 100uSec
; DELAY_value	= (500uSec - 4) / 3
;		= 165.33
;		= 165
;===========================================================================
WAIT_A_SEC	
		call WAIT_500MS
		call WAIT_500MS
		
WAIT_750MS	
		call WAIT_256MS
		call WAIT_256MS
		call WAIT_256MS
		
WAIT_500MS	
		call WAIT_128MS
		call WAIT_128MS
		call WAIT_128MS
		call WAIT_128MS


WAIT_256MS	
		call WAIT_128MS
		call WAIT_128MS
		
WAIT_128MS	
		movlw	D'240'		; 128 msec
		movwf	X_DLY_VAL
LOOP_128MS	
		call	DELAY		;WAIT 500uSec
		decfsz	X_DLY_VAL, F
		goto	LOOP_128MS
		return
		
WAIT_64MS	
		movlw	D'120'		; 64 msec
		movwf	X_DLY_VAL	;
LOOP_64MS	
		call	DELAY		;WAIT 500uSec
		decfsz	X_DLY_VAL, F
		goto	LOOP_64MS	
		return
		
WAIT_32MS	
		movlw	D'60'		; 30 msec
		movwF	X_DLY_VAL
LOOP_32MS	
		call	DELAY		; WAIT 500uSec
		decfsz	X_DLY_VAL, F
		goto	LOOP_32MS
		return


WAIT_15MS	
		movlw	D'30'		; 15 msec
		movwf	X_DLY_VAL
LOOP_15MS	
		call	DELAY		;wait 500uSec
		decfsz	X_DLY_VAL, F
		goto	LOOP_15MS
		return

;DELAY ROUTINE = 500uSec OR 1/2 A MILISECOND.
DELAY		
		movlw	H'FF'		; +1		1 CYCLE
		movwF	DELAY_VAL	; +2		1 CYCLE
		
DELAY_LOOP	
		decfsz	DELAY_VAL, F; step 1	1 CYCLE
		goto	DELAY_LOOP	; step 2	2 CYCLE
		return			; +3		2 CYCLE
		
		end
